
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <assert.h>

#include "genquery.h"

extern long M;		// = 0xffff;
extern long BSIZE;	// = 1000;

/*****************************************************************************
double random_gaussian(double mu, double sigma)
INPUTS: 
	mu - mean
	sigma - SD

OUTPUT:
	random double from a gaussian distribution with SD=sigma and mean=mu

GLOBALS:
	NIL

Function:
	Return a random double from a gaussian distribution with SD=sigma and 
	mean=mu
*****************************************************************************/
double random_gaussian(double mu, double sigma)
{
	static int HAS_OTHER=FALSE;
	static double other;
	double r1, r2, rsqr, p;

	if (HAS_OTHER) {
		HAS_OTHER = FALSE;
		return(other*sigma + mu);
	} else {
		do {
			r1 = 2.0 * (random_val(0.0, 1.0)) - 1.0;
			r2 = 2.0 * (random_val(0.0, 1.0)) - 1.0;
			rsqr = r1*r1 + r2*r2;
		} while (rsqr >= 1.0 || rsqr == 0.0);
		
		p=sqrt(-2.0*log(rsqr)/rsqr);
		
		other=r2*p;
		HAS_OTHER=TRUE;
		
		return (r1*p*sigma + mu);
	}
}

/*
	N = current transaction
	t1, t2 = holders for tqmin and tqmax
	item = holder for item to search

	Generate query of the form - find frequency of "item" in 
	interval ["tqmin", "tqmax"]
*/
void gen_query_single(long N, long *t1, long *t2, long *item)
{
	long mu, sigma;
	
	mu = N - N/10;
	sigma = N/8;

	do {	// pick t2 from a gaussian distribution
		*t2 = (long)random_gaussian(mu, sigma);
	} while (*t2 > N);
	
	do {	// pick t1 from a uniform distribution
		*t1 = (long)(*t2 - BSIZE * random_val(1.0, 20.0));
	} while ((*t1 < 0) || (*t1 > *t2));

	*item = (long)random_val(1, M);
}


/*
	N = current transaction
	t1, t2 = holders for tqmin and tqmax
	thresh = holder for threshold (0.0 - 1.0)

	Generate query of the form - find all items having threshold "thresh" or 
	higher in interval ["tqmin", "tqmax"]
*/
void gen_query(long N, long *t1, long *t2, double *thresh, int width)
{
	long mu, sigma;
	
	mu = N - N/10;
	sigma = N/8;

	//printf("gen_query");

	do {	// pick t2 from a gaussian distribution
		*t2 = (long)random_gaussian(mu, sigma);
	} while (*t2 > N);
	
	*t1 = *t2 - width * BSIZE;
	if (*t1 < 0) {
		*t1 = 1;
	}
	
        assert(t1 > 0);
        assert(t2 > 0);

	/*
	do {	// pick t1 from a uniform distribution
		*t1 = (long)(*t2 - BSIZE * random_val(1.0, 8.0));
	} while ((*t1 < 0) || (*t1 > *t2));
	*/

	*thresh = random_val(0.002, 0.01);
}

#ifdef TEST
int main()
{
	long N;
	long t1, t2, item;
	double th;

        //srand(time(NULL));

	for (N=10000; N<100000; N++) {
		if (N % 5000 == 0) {
			//gen_query_single(N, &t1, &t2, &item);
			//printf("\nN=%ld: (%ld, %ld) (%ld)", N, t1, t2, item); 
			gen_query(N, &t1, &t2, &th);
			printf("\nN=%ld: (%ld, %ld) (%f)", N, t1, t2, th); 
		}
	}

	return(0);
}

#endif
